// ***************************************************************
// Name:      CLUSThack.c
// Purpose:   This module was developed in the event, 
// 	      	ClusterHack: Efficient OTU clustering, alignment and noise 
// 	      	removal for short read amplicons, August 27th-29th, University of Glasgow
// 	      attended by, 
// 	      Umer Zeeshan Ijaz (Umer.Ijaz@glasgow.ac.uk)
// 	      Jiajie Zhang (bestzhangjiajie@gmail.com)
// 	      Christopher Quince (quince@civil.gla.ac.uk)
// 	      Frederic Mahe (mahe@rhrk.uni-kl.de) 
// 	      Torbjorn Rognes (torognes@ifi.uio.no)
//
// 	      The purpose of this module is to provide fast multi-threaded algorithms
// 	      for filtering and alignment of sequences.
// 	      So far there is support for fast q-gram based filtering.
// 	      This work is still in progress
//
// Version:   0.2 qgram function now returns distance matrix
// History:   0.1 basic q-gram based filtering incorporated
// Authors:   Umer Zeeshan Ijaz (Umer.Ijaz@glasgow.ac.uk)
//                 http://userweb.eng.gla.ac.uk/umer.ijaz
//            Torbjorn Rognes (fast q-gram code contributed)     
// Last Modified:   2013-09-16
// License:   Copyright (c) 2013 Computational Microbial Genomics Group, University of Glasgow, UK
//
//            This program is free software: you can redistribute it and/or modify
//            it under the terms of the GNU General Public License as published by
//            the Free Software Foundation, either version 3 of the License, or
//            (at your option) any later version.
//
//            This program is distributed in the hope that it will be useful,
//            but WITHOUT ANY WARRANTY; without even the implied warranty of
//            MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//            GNU General Public License for more details.
//
//            You should have received a copy of the GNU General Public License
//            along with this program.  If not, see <http://www.gnu.org/licenses/>.
// **************************************************************


#include <Python.h>
#include <pthread.h>

// === qgram functions and imports === //

#define QGRAMLENGTH 5
#define QGRAMVECTORBITS (1<<(2*QGRAMLENGTH))
#define QGRAMVECTORBYTES (QGRAMVECTORBITS/8)
#define MAX_THREADS 256

char map_nt[256] =
  {
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, 1, -1, 2, -1, -1, -1, 3, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, 4, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, 1, -1, 2, -1, -1, -1, 3, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, 4, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1
  };

long threads=8;
char *qgram_vectors;

void fatal(const char * msg)
{
  fprintf(stderr, "Error: %s\n", msg);
  exit(1);
}



inline unsigned char * db_getqgramvector(unsigned long seqno)
{
  return qgram_vectors+seqno*QGRAMVECTORBYTES;
}



void printqgrams(unsigned char * qgramvector)
{
  /* print qgramvector */
  fprintf(stderr, "qgram vector:\n");
  for(int i = 0; i < QGRAMVECTORBYTES; i++)
  {
    fprintf(stderr, "%02x", qgramvector[i]);
    if ((i % 32) == 31)
      fprintf(stderr, "\n");
  }
}

void findqgrams(char * seq, unsigned long seqlen,unsigned char * qgramvector)
{
  /* set qgram bit vector by xoring occurrences of qgrams in sequence */

  memset(qgramvector, 0, QGRAMVECTORBYTES);
  
  unsigned long qgram = 0;
  unsigned long i = 0;

  while((i < QGRAMLENGTH-1) && (i<seqlen))
  {
    qgram = (qgram << 2) | (map_nt[seq[i]] - 1);
    i++;
  }

  while(i < seqlen)
  {
    qgram = (qgram << 2) | (map_nt[seq[i]] - 1);
    qgramvector[(qgram >> 3) & (QGRAMVECTORBYTES-1)] ^= (1 << (qgram & 7));
    i++;
  }
}

/*
 * Unable to get the Mac gcc compiler v 4.2.1 produce the real
 * popcnt instruction. Therefore resorting to assembly code.
 * */

#define popcnt_asm(x,y) \
__asm__ __volatile__ ("popcnt %1,%0" : "=r"(y) : "r"(x));

inline unsigned long popcount(unsigned long x)
{
  unsigned long y;
  popcnt_asm(x,y);
  return y;
}

unsigned long compareqgramvectors(unsigned char * a, unsigned char * b)
{
  /* count number of different bits */
  unsigned long count = 0;

#if 1

  /* using ordinary 64-bit registers */

  unsigned long *ap = (unsigned long*)a;
  unsigned long *bp = (unsigned long*)b;

#if 1

  /* with ordinary loop */

  for(int i = 0; i < QGRAMVECTORBYTES/8; i++)
    count += popcount(ap[i] ^ bp[i]);

#else

  /* unrolled loop */

  count += popcount(ap[ 0] ^ bp[ 0]);
  count += popcount(ap[ 1] ^ bp[ 1]);
  count += popcount(ap[ 2] ^ bp[ 2]);
  count += popcount(ap[ 3] ^ bp[ 3]);
  count += popcount(ap[ 4] ^ bp[ 4]);
  count += popcount(ap[ 5] ^ bp[ 5]);
  count += popcount(ap[ 6] ^ bp[ 6]);
  count += popcount(ap[ 7] ^ bp[ 7]);
  count += popcount(ap[ 8] ^ bp[ 8]);
  count += popcount(ap[ 9] ^ bp[ 9]);
  count += popcount(ap[10] ^ bp[10]);
  count += popcount(ap[11] ^ bp[11]);
  count += popcount(ap[12] ^ bp[12]);
  count += popcount(ap[13] ^ bp[13]);
  count += popcount(ap[14] ^ bp[14]);
  count += popcount(ap[15] ^ bp[15]);

#endif

#else

  /* SSE2 version using the 128-bit XMM registers */

  __m128i * ap = (__m128i*) a;
  __m128i * bp = (__m128i*) b;

  for(int i = 0; i < 8; i++)
    {
      __m128i x = _mm_load_si128(ap+i);
      x = _mm_xor_si128(x, bp[i]);
      count += popcount(_mm_extract_epi64(x, 0));
      count += popcount(_mm_extract_epi64(x, 1));
    }

#endif

  return count;
}

inline unsigned long qgram_diff(unsigned long a, unsigned long b)
{
  unsigned long diffqgrams = compareqgramvectors(db_getqgramvector(a),
db_getqgramvector(b));
  unsigned long mindiff = (diffqgrams + 2*QGRAMLENGTH - 1)/(2*QGRAMLENGTH);
  return mindiff;
}

struct thread_info_struct
{
  pthread_t id;
  unsigned long seed;
  unsigned long listlen;
  unsigned long * amplist;
  unsigned long * difflist;
} thread_info[MAX_THREADS];

void * qgram_diff_worker(void * vp)
{
  long t = (long) vp;

  unsigned long seed = thread_info[t].seed;
  unsigned long listlen = thread_info[t].listlen;
  unsigned long * amplist = thread_info[t].amplist;
  unsigned long * difflist = thread_info[t].difflist;

  for(unsigned long i=0; i<listlen; i++)
    difflist[i] = qgram_diff(seed, amplist[i]);
  return 0;
}

void qgram_diff_parallel(unsigned long seed,
unsigned long listlen,
unsigned long * amplist,
unsigned long * difflist)
{
  long thr = threads;
  
  const unsigned long m = 3000;

  if (listlen < m*thr)
    thr = (listlen+m-1)/m;
  
  if (thr == 1)
    {
      for(unsigned long i=0; i<listlen; i++)
difflist[i] = qgram_diff(seed, amplist[i]);
    }
  else
    {
      long t;
      void * status;
      
      unsigned long * next_amplist = amplist;
      unsigned long * next_difflist = difflist;
      unsigned long listrest = listlen;
      unsigned long thrrest = thr;
      unsigned long chunk;
      
      for(t=0; t<thr; t++)
{
thread_info[t].seed = seed;
thread_info[t].amplist = next_amplist;
thread_info[t].difflist = next_difflist;
chunk = (listrest + thrrest - 1) / thrrest;
thread_info[t].listlen = chunk;
next_amplist += chunk;
next_difflist += chunk;
listrest -= chunk;
thrrest--;
}

      for(t=0; t<thr; t++)
if (pthread_create(&thread_info[t].id, 0, qgram_diff_worker, (void *)t))
fatal("Cannot create thread.");
      
      for(t=0; t<thr; t++) {
if (pthread_join(thread_info[t].id, &status))
fatal("Cannot join thread.");
      }
    }
}
// =================================== //





 // === helper functions to get data from python === //
static char**
get_char_array(PyObject* data, int* data_size)
{
    int i, size;
    char** out;
    PyObject* seq;

    seq = PySequence_Fast(data, "expected a sequence");
    if (!seq)
        return NULL;

    size = PySequence_Size(seq);
    if (size < 0)
        return NULL;

    if (data_size)
        *data_size = size;

    out = (char**) PyMem_Malloc((size+1) * sizeof(char *));
    if (!out) {
        Py_DECREF(seq);
        PyErr_NoMemory();
        return NULL;
    }

    for (i = 0; i < size; i++)
        out[i] = PyString_AsString(PySequence_Fast_GET_ITEM(seq, i));

    Py_DECREF(seq);

    if (PyErr_Occurred()) {
        PyMem_Free(out);
        out = NULL;
    }

    return out;
}

static unsigned long*
get_long_array(PyObject* data, int* data_size)
{
    int i, size;
    unsigned long* out;
    PyObject* seq;

    seq = PySequence_Fast(data, "expected a sequence");
    if (!seq)
        return NULL;

    size = PySequence_Size(seq);
    if (size < 0)
        return NULL;

    if (data_size)
        *data_size = size;

    out = (unsigned long*) PyMem_Malloc(size * sizeof(unsigned long));
    if (!out) {
        Py_DECREF(seq);
        PyErr_NoMemory();
        return NULL;
    }

    for (i = 0; i < size; i++)
        out[i] = PyInt_AsLong(PySequence_Fast_GET_ITEM(seq, i));

    Py_DECREF(seq);

    if (PyErr_Occurred()) {
        PyMem_Free(out);
        out = NULL;
    }

    return out;
}


 // ================================================ //


static PyObject* qgram(PyObject* self, PyObject* args)
{
    // === import data from python list === //
    int i,j;
    
    int vals_size=0; //size of the supplied sequences
    int filtered_list_size=0; //size of the sequence indices for which distance should be generated	 
    int threshold=10; //threshold used for distance-based filtered
    int format=1; //format used


    PyObject* vals_in;
    PyObject* filtered_list_in;	
    if (!PyArg_ParseTuple(args, "OiiiO:qgram",&vals_in,&threshold,&format,&threads,&filtered_list_in))
	return NULL;
    char** vals;
    unsigned long* filtered_list;
    //extract sequence data, store it in vals_in with number of sequences in vals_size
    vals=get_char_array(vals_in,&vals_size);

    //extract filtered list, store it in filtered_list_in with total members in filtered_list_size	

    unsigned long *difflist;

    if (filtered_list_in==Py_None){
	filtered_list=(unsigned long *) PyMem_Malloc((vals_size) * sizeof(unsigned long));
	filtered_list_size=vals_size;
        for(j=0;j<vals_size;j++){
                filtered_list[j]=j;
                }
	difflist=(unsigned long *) PyMem_Malloc((vals_size) * sizeof(unsigned long));
	}
    else{
    	filtered_list=get_long_array(filtered_list_in,&filtered_list_size);
	difflist=(unsigned long *) PyMem_Malloc((filtered_list_size) * sizeof(unsigned long));
    	/*for(j=0;j<filtered_list_size;j++){
        	printf("filtered_list[%d]=%lu\n",j,filtered_list[j]);
    		}
	*/
	}

    // == generate qgram vectors == //		
    qgram_vectors  = (char*) PyMem_Malloc((vals_size) * QGRAMVECTORBYTES);
    for(i=0;i<vals_size;i++){
	//printf("vals[%d]=%s\n",i,vals[i]);
	findqgrams(vals[i], strlen(vals[i]),qgram_vectors+i*QGRAMVECTORBYTES);
	//printqgrams(qgram_vectors+i*QGRAMVECTORBYTES);
	}
    // ============================ // 
  
    PyObject *pylist;
    PyObject *pysublist;
    pylist=PyList_New(vals_size);

    // ==================================== //  
    for(i=0;i<vals_size;i++){
        qgram_diff_parallel(i,(unsigned long)filtered_list_size,filtered_list,difflist);
        pysublist=PyList_New(0);
        for(j=0;j<filtered_list_size;j++){
                if ((int)format==1) // return assignments
                        {
                        if((difflist[j]<=(unsigned long)threshold)&& (i!=filtered_list[j]))
                                {
                                PyList_Append(pysublist,Py_BuildValue("i",filtered_list[j]));
                                }
                        }
                else if((int)format==2)
                        {
                        PyList_Append(pysublist,Py_BuildValue("i",difflist[j]));        
                        //printf("%lu\t",difflist[j]);
                        }       
                }
        PyList_SetItem(pylist,i,pysublist);
        //printf("\n");
        }




    PyMem_Free(filtered_list);
    PyMem_Free(difflist);
    PyMem_Free(qgram_vectors);  
    PyMem_Free(vals);


    return pylist;	



/*

 
    for(i=0;i<filtered_list_size;i++){
    	printf("filtered_list[%d]=%lu\n",i,filtered_list[i]);
    }		
    int nSubsetCount;	
    for(i=0;i<vals_size;i++){
	nSubsetCount=0;
	for(j=0;j<filtered_list_size;j++)
		{
		
		nSubsetCount++;
		}
    }	



   

    unsigned long *amplist=(unsigned long *) PyMem_Malloc((vals_size) * sizeof(unsigned long));
    unsigned long *difflist=(unsigned long *) PyMem_Malloc((vals_size) * sizeof(unsigned long));


    PyObject *pylist;
    PyObject *pysublist;
    pylist=PyList_New(vals_size);

    // ==================================== //	
    for(j=0;j<vals_size;j++){
                amplist[j]=j;
        }
    for(i=0;i<vals_size;i++){
	qgram_diff_parallel(i,(unsigned long)vals_size,amplist,difflist);
	pysublist=PyList_New(0);
	for(j=0;j<vals_size;j++){
		if ((int)format==1) // return assignments
			{
			if((difflist[j]<=(unsigned long)threshold)&& (i!=j))
				{
				PyList_Append(pysublist,Py_BuildValue("i",j));
				}
			}
		else if((int)format==2)
			{
			if(j<=i)// return ragged distance matrix
				{
				if(i==0&&j==0)
					pysublist=PyList_New(0);
				else
					if(i!=j)
						PyList_Append(pysublist,Py_BuildValue("i",difflist[j]));	
						//printf("%lu\t",difflist[j]);
				}
			}	
		}
	PyList_SetItem(pylist,i,pysublist);
	//printf("\n");
	}
   	
	 
	







    //PyObject *pylist;
    //pylist=PyTuple_New(2);
    //PyTuple_SetItem(pylist,0,PyString_FromString(alignment[0]));
    //PyTuple_SetItem(pylist,1,PyString_FromString(alignment[1]));	
    PyMem_Free(amplist);
    PyMem_Free(difflist);
    PyMem_Free(qgram_vectors);	
    PyMem_Free(vals);
    PyMem_Free(filtered_list);

    return pylist; 
    //Py_RETURN_NONE;

*/
}
 
static PyMethodDef CLUSTHACKMethods[] =
{
     {"qgram", qgram, METH_VARARGS, "qgram"},
     {NULL, NULL, 0, NULL}
};
 
PyMODINIT_FUNC
 
initCLUSThack(void)
{
     (void) Py_InitModule("CLUSThack", CLUSTHACKMethods);
}
